/***********************************************************************
                        TQ2PISTONENGINES
           Using ArduinoJoystickLibrary by Matthew Heironimus
          https://github.com/MHeironimus/ArduinoJoystickLibrary
 **********************************************************************/
#include "EEPROM.h"

#define DEVICE_ID  5

#define ANALOG_FIFO_LENGTH (12)
#define CHANNELS_NUMBER  (6)

#define OPERATION_FLAG_SHOW_RAW_VALUES   1
#define OPERATION_FLAG_SHOW_SCALED       2
#define OPERATION_FLAG_SHOW_ERROR        4
#define OPERATION_FLAG_SHOW_CLEAN        8

short error_left[128];
short error_right[128];


enum {
    REVERSE_LEVER_DOWN,
    REVERSE_LEVER_CENTER,
    REVERSE_LEVER_UP
};

typedef struct {
   char  state;
   short value;
} REVERSE_LEVER;

REVERSE_LEVER reverse_lever[2] = {{0,0},{0,0}};


void reverse_lever_fsm(void)
{
    int i;
    for(i=0;i<2;i++)
    {  
      switch(reverse_lever[i].state)
      {
         case REVERSE_LEVER_DOWN:
            if(reverse_lever[i].value > 0)
                    reverse_lever[i].value--;
         break;
         
         case REVERSE_LEVER_CENTER:
         
             if(reverse_lever[i].value < 1024)
                   reverse_lever[i].value++;
             if(reverse_lever[i].value > 1024)
                   reverse_lever[i].value--;
         break;
         
         case REVERSE_LEVER_UP:
             if(reverse_lever[i].value < 2047)
                   reverse_lever[i].value++;
         break;
      }
    }  
}


int operation_flags = 0;

void moving_average_filter_init(void);
short moving_average_filter(int ch, short in);
void mlx90316_setup(void);
void read_position(int *pos_pull) ;
void discrete_task_init(void);
void discrete_task( void);
void load_config(void);
void save_config(void);

static short pos0[4] = {593,  9275, 5891, 9873};
static short pos1[4] = {5303, 5838, 9600, 5814};
static short pos_pull[4];
static short offset[4] = {0, 0, 0, 0};
static short delta[4] = {0, 0, 0, 0};
static long factor[4] = {0, 0, 0, 0};

#include "Joystick.h"
/*--------------------------------------------------------
  "THE BEER-WARE LICENSE" (Revision 42):
  Alex Kostyuk wrote this code. As long as you retain this
  notice you can do whatever you want with this stuff.
  If we meet some day, and you think this stuff is worth it,
  you can buy me a beer in return.
  ----------------------------------------------------------*/
#if 0
Joystick_::Joystick_(
  uint8_t hidReportId,
  uint8_t joystickType,
  uint8_t buttonCount,
  uint8_t hatSwitchCount,
  bool includeXAxis,
  bool includeYAxis,
  bool includeZAxis,
  bool includeRxAxis,
  bool includeRyAxis,
  bool includeRzAxis,
  bool includeRudder,
  bool includeThrottle,
  bool includeAccelerator,
  bool includeBrake,
  bool includeSteering)
#endif

#define STX   '{'
#define ETX   '}'

const char hex_table[] = "0123456789ABCDEF";

static unsigned char nibble(int hex)
{
  if (hex >= '0' && hex <= '9') return (hex - '0');
  if (hex >= 'A' && hex <= 'F') return (hex - 'A' + 10);
  if (hex >= 'a' && hex <= 'f') return (hex - 'a' + 10);
  return 0;
}

static int hex2int(char *hh)
{
  int nib1 = nibble(hh[0] & 0xFF);
  int nib2 = nibble(hh[1] & 0xFF);
  return  (nib1 << 4) + nib2;
}

unsigned long hex2ulong(const char *in)
{
  unsigned long ul  = 0;
  char *wp = (char*)in;
  int i;

  for (i = 0; i < 4; i++)
  {
    ul <<= 8;
    ul += hex2int(wp);
    wp += 2;
  }

  return ul;
}




Joystick_ Joystick(JOYSTICK_DEFAULT_REPORT_ID,
                   JOYSTICK_TYPE_MULTI_AXIS, 32, 0,
                   true, /* XAxis       */
                   true, /* YAxis       */
                   true, /* ZAxis       */
                   true, /* RxAxis      */
                   true, /* RyAxis      */
                   true, /* RzAxis      */
                   false,  /* Rudder      */
                   false,  /* Throttle    */
                   false, /* Accelerator */
                   false, /* Brake       */
                   false  /* Steering    */
                  );


void factor_calculation(void)
{
  int i;
  long tmp;
  for(i=0;i<4;i++)
  {
    offset[i] = pos0[i];
    delta[i] = pos1[i] - pos0[i]; 
    tmp = delta[i];
    tmp *= 100000;
    tmp = tmp / 2048;
    factor[i] = tmp;
  }
 
}


void setup()
{
  load_config();
  factor_calculation();
  Serial.begin(115200);
  moving_average_filter_init();
  mlx90316_setup();
  discrete_task_init();

  Joystick.setXAxisRange(0, 2048);
  Joystick.setYAxisRange(0, 2048);
  Joystick.setZAxisRange(0, 2048);
  Joystick.setRxAxisRange(0, 2048);
  Joystick.setRyAxisRange(0, 2048);
  Joystick.setRzAxisRange(0, 2048);

  Joystick.begin(false);
}

#if 1

#define COMMAND_BUFFER_SIZE 100
char command_buffer[COMMAND_BUFFER_SIZE+1];
char command_buffer_ndx = 0;

void serial_read(void)
{
  static char buf[8];
  static int ndx = 0;
  int itmp;
  if (Serial.available() > 0)
  {
    char ch = (char)Serial.read();
    switch (ch)
    {
      case STX:
        ndx = 0;
        break;

      case ETX:

        if (buf[0] == '7')
        {
          buf[7] = 0;
          //trimmer_position = atoi(&buf[3]);

        }

        if (buf[0] == 'D')
        {
          Serial.write('!');
          Serial.write(STX);
          Serial.write('F');
          itmp = DEVICE_ID;
          Serial.write(hex_table[itmp >> 4]);
          Serial.write(hex_table[itmp & 0x0F]);
          Serial.write(ETX);

        }
        else
          Serial.write('!');
        break;
      default:
        if (ndx < 8)
          buf[ndx++] = ch;
        
        if(command_buffer_ndx < COMMAND_BUFFER_SIZE)
          command_buffer[command_buffer_ndx++] = ch;
        if(ch == 0x0D || ch == 0x0A)
        {
           command_buffer[command_buffer_ndx] = 0;
           command_buffer_ndx = 0;
           command_com(command_buffer);
        }
    }
  }

}
#endif

long scale(int i)
{
#if 0
      long tmp = pos_pull[i];
      return tmp;
#else      
      long tmp = (pos_pull[i] - offset[i]);
      tmp *= 100000;
      tmp = tmp / factor[i];
      return tmp;
#endif      
}

short addr_limit(short cur_val)
{
    cur_val >>= 4;
    if(cur_val > 127)
       cur_val = 127;
    if(cur_val<0)
       cur_val = 0;
    return cur_val;      
}

void loop(void)
{
  static unsigned long ms_reftime = 0;
  static short cur[6] = {0, 0, 0, 0, 0, 0};
  unsigned long cur_time = millis();
  static char cnt;
  static int timer = 100;
  serial_read();

  if ((cur_time - ms_reftime) >= 1)
  {
    ms_reftime = cur_time;
    reverse_lever_fsm();
    if (++cnt == 10)
    {
      discrete_task();
      read_position(pos_pull);
      cnt = 0;

      //short tmp = 
      //reverse_lever[0].state || 
      //reverse_lever[1].state;
      

       cur[0] = scale(0);
        

      cur[1] = scale(1);
      cur[2] = scale(2);
      cur[3] = scale(3);
      cur[4] = reverse_lever[0].value;
      cur[5] = reverse_lever[1].value;;

      if(operation_flags & OPERATION_FLAG_SHOW_SCALED)
      {
        int i;
        for(i=0;i<4;i++)
        {
          Serial.print(cur[i]);
          Serial.print(" ");
        }
        Serial.println(" ");
      }

      if(operation_flags & OPERATION_FLAG_SHOW_RAW_VALUES)
      {
        int i;
        for(i=0;i<4;i++)
        {
          Serial.print(pos_pull[i]);
          Serial.print(" ");
        }
        Serial.println(" ");
      }

      if(operation_flags & OPERATION_FLAG_SHOW_ERROR)
      {
         short aver = (cur[1] + cur[2])/2;
         short dleft = cur[1] - aver; 
         short dright = cur[2] - aver; 
         error_left[addr_limit(cur[1])] = dleft;
         error_right[addr_limit(cur[2])] = dright;    
         Serial.print(addr_limit(cur[1]));
         Serial.print(" ");
         Serial.print(addr_limit(cur[2]));
         Serial.print(" ");
         Serial.print(aver);
         Serial.print(" ");
         Serial.print(dleft);
         Serial.print(" ");
         Serial.println(dright);
      }

     cur[1] -= error_left[addr_limit(cur[1])];
     cur[2] -= error_right[addr_limit(cur[2])];
      
      if(operation_flags & OPERATION_FLAG_SHOW_CLEAN)
      {
        int i;
        for(i=0;i<4;i++)
        {
          Serial.print(cur[i]);
          Serial.print(" ");
        }
        Serial.println(" ");

      }

    }

    Joystick.setXAxis(cur[0]);
    Joystick.setYAxis(cur[1]);
    Joystick.setZAxis(cur[2]);
    Joystick.setRxAxis(cur[3]);
    Joystick.setRyAxis(cur[4]);
    Joystick.setRzAxis(cur[5]);

    Joystick.sendState();
  }
}

typedef struct
{
  int   push_indx;
  int   pop_indx;
  short buff[ANALOG_FIFO_LENGTH];
} SIMPLE_FIFO;

SIMPLE_FIFO analog_fifo[CHANNELS_NUMBER];

short fifo_push(SIMPLE_FIFO *fifo, short data);
short fifo_pop(SIMPLE_FIFO *fifo, short *data);

void moving_average_filter_init(void)
{
  memset(analog_fifo, 0, sizeof(analog_fifo));
}



short moving_average_filter(int ch, short in)
{
  static short sum[CHANNELS_NUMBER] = {0, 0, 0, 0, 0, 0};
  short tmp = in;
  short prev;
  sum[ch] += in;
  if (!fifo_push(&analog_fifo[ch], in))
  {
    fifo_pop(&analog_fifo[ch], &prev);
    fifo_push(&analog_fifo[ch], in);
    sum[ch] -= prev;
    tmp = (sum[ch] / (ANALOG_FIFO_LENGTH - 1));
    return tmp;
  }


  return 0;
}

short fifo_push(SIMPLE_FIFO *fifo, short data)
{
  int indx;
  indx = fifo->push_indx + 1;
  indx = indx % ANALOG_FIFO_LENGTH;

  if (indx == fifo->pop_indx)
    return 0; // fifo is full
  fifo->buff[indx] = data;
  fifo->push_indx = indx;
  return 1;
}

short fifo_pop(SIMPLE_FIFO *fifo, short *data)
{
  int indx;
  if (fifo->push_indx == fifo->pop_indx)
    return 0; // fifo is empty
  indx = fifo->pop_indx + 1;
  indx = indx % ANALOG_FIFO_LENGTH;
  *data = fifo->buff[indx];
  fifo->pop_indx = indx;
  return 1;
}

#define MLX_SCK      16
#define MLX_SS       15
#define MLX_SIO_PORT_DIR DDRF
#define MLX_SIO_PORT_IN  PINF
#define MLX_SIO_PORT_OUT PORTF
#define MLX_SIO_PORT_MASK 0xF0


void mlx90316_setup(void)
{
  MLX_SIO_PORT_OUT  |= MLX_SIO_PORT_MASK;
  MLX_SIO_PORT_DIR  &= ~MLX_SIO_PORT_MASK;
  pinMode(MLX_SCK , OUTPUT);
  pinMode(MLX_SS , OUTPUT);
}

unsigned long ref = 0;

#if 0
void loop()
{
  int pos_pull[4];
  int i;

  if ((millis() - ref) > 7)
  {
    ref = millis();
    read_position(pos_pull) ;
    for (i = 0; i < 4; i++)
    {

      Serial.print(pos_pull[i]); Serial.print(" ");
    }
    Serial.println("");
  }
}
#endif

void spi_ini_sequence(void)
{
  unsigned int mask = 1;
  int i, j;
  MLX_SIO_PORT_DIR = MLX_SIO_PORT_MASK;

  for (j = 0; j < 16; j++)
  {
    int i;
    byte out = 0;
    for (i = 0; i < 4; i++)
    {
      out = out << 1;
      if (0xFF55 & mask)
        out |=  0x10;
    }
    out = (MLX_SIO_PORT_OUT & ~MLX_SIO_PORT_MASK) | out; 
    MLX_SIO_PORT_OUT = out;
    
    mask = mask << 1;


    digitalWrite(MLX_SCK, 1);  // ___/
    digitalWrite(MLX_SCK, 0);  //      \________

    if(j==7)
    {
      delayMicroseconds(5);
    }
    
  }
  delayMicroseconds(5);

  MLX_SIO_PORT_OUT |= MLX_SIO_PORT_MASK;
  MLX_SIO_PORT_DIR &= ~MLX_SIO_PORT_MASK;
}


void spi_write_16uint(uint16_t *pull)
{
  unsigned int mask = 1;
  int i, j;
  MLX_SIO_PORT_DIR |= MLX_SIO_PORT_MASK;;

  for (j = 0; j < 16; j++)
  {
    int i;
    byte out = 0;
    for (i = 0; i < 4; i++)
    {
      out = out << 0x10;
      if (pull[i] & mask)
        out |=  1;
    }
    out = (MLX_SIO_PORT_OUT & MLX_SIO_PORT_MASK) | out; 
    MLX_SIO_PORT_OUT = out;

    
    mask = mask << 1;


    digitalWrite(MLX_SCK, 1);  // clk _____/
    digitalWrite(MLX_SCK, 0);  // clk        \____

    if(j==7)
    {
      delayMicroseconds(5);
    }
    
  }
  delayMicroseconds(5);

  MLX_SIO_PORT_OUT |= MLX_SIO_PORT_MASK;
  MLX_SIO_PORT_DIR &= ~MLX_SIO_PORT_MASK;
}

//*************************************************************

void spi_read_16uint(uint16_t *pull)
{

  int i, j;

  for (j = 0; j < 16; j++)
  { 

    digitalWrite(MLX_SCK, 1);   // clk ____/   

    byte mask = 0x10;
    byte in = MLX_SIO_PORT_IN;
    for (i = 0; i < 4; i++)
    {
      pull[i] = (pull[i] << 1);
      if (in & mask)
        pull[i] |= 1;
      mask = mask << 1;
    }
    
    digitalWrite(MLX_SCK, 0);  //  clk  \_______
    if(j==7)
    {
      delayMicroseconds(5);
    }
  }
  delayMicroseconds(5);
}

void read_position(short *pos_pull)
{

  int i;
  
  digitalWrite(MLX_SS , 0);

  unsigned int pos_pull_dir[8];
  unsigned int pos_pull_inv[8];
  unsigned int dummy[8];

  spi_ini_sequence();
  spi_read_16uint(pos_pull_dir);
  spi_read_16uint(pos_pull_inv);
  spi_read_16uint(dummy);
  spi_read_16uint(dummy);
  
  for (i = 0; i < 4; i++)
  {
    if ( (pos_pull_dir[i] + pos_pull_inv[i] + 1) != 0 )
      pos_pull[i] = -1;
    else
    {
      if ((pos_pull_dir[i] & 3) == 2)
      {
        if ( pos_pull_dir[i] & (1 << 4))
          pos_pull[i] = -2; // too strong
        if ( pos_pull_dir[i] & (1 << 5))
          pos_pull[i] = -3; // too weak
      }
      else if ((pos_pull_dir[i] & 3) == 1)
      {
        pos_pull[3-i] = pos_pull_dir[i] >> 2;
      }
    }

  }

  
  digitalWrite(MLX_SS , 1);

}


//-------------------------------------------
//               Discrete task
//-------------------------------------------
#define BITS_MAXIMUM   12 /*ITEMS_MAX*/
int shift = 0;

#define FLAG_INVERTED   (1<<0)
#define FLAG_ENABLED    (1<<1)

enum {
  PUSH_BUTTON = 0,
  TOGGLE_SWITCH,
  ROTARY_SWITCH_POLE,
  ENCODER_A,
  ENCODER_B,
  SHIFT
};

typedef struct {
int counter;
char signal;
int  max;
char evcode;
char flag;
char mode;
char pin;
char emitted_code_rising;
char emitted_code_falling;
int pulse_timer;
int pulse_timer_value;
} BITS_DESCRIPTION;

BITS_DESCRIPTION  bits[BITS_MAXIMUM];
 
#define input_init(x,y,z,m,r,f, t)   do {\
bits[x].counter = 0;\
bits[x].signal = 0;\
bits[x].max = y;\
bits[x].flag = (FLAG_ENABLED | FLAG_INVERTED);\
bits[x].pin = z;\
pinMode(z,INPUT_PULLUP);\
bits[x].mode = m;\
bits[x].emitted_code_rising = r;\
bits[x].emitted_code_falling = f;\
bits[x].pulse_timer_value = t;\
} while(0)

void discrete_task_init(void)
{
  int i;
  memset(bits,0,sizeof(bits));
#define DEBOUNCE_DELAY   10

//input_init(0,DEBOUNCE_DELAY, 0, ROTARY_SWITCH_POLE,0,-1,   0);
//input_init(1,DEBOUNCE_DELAY,0,  ROTARY_SWITCH_POLE, 1,-1,  0);
//input_init(2,DEBOUNCE_DELAY,0,  ROTARY_SWITCH_POLE, 2,-1,  0);
//input_init(3,DEBOUNCE_DELAY, 0, ROTARY_SWITCH_POLE,3,-1,   0);
//input_init(4,DEBOUNCE_DELAY,0,  ROTARY_SWITCH_POLE, 4,-1,  0);
//input_init(5,DEBOUNCE_DELAY,0,  ROTARY_SWITCH_POLE, 5,-1,  0);
//input_init(6,DEBOUNCE_DELAY,0,  ROTARY_SWITCH_POLE, 6,-1,  0);
//input_init(7,DEBOUNCE_DELAY,0,  ROTARY_SWITCH_POLE, 7,-1,  0);

input_init(0,DEBOUNCE_DELAY,0,   PUSH_BUTTON,   0, -1, 0);
input_init(1,DEBOUNCE_DELAY,1,   PUSH_BUTTON,   1, -1, 0);
input_init(2,DEBOUNCE_DELAY,2,   PUSH_BUTTON,   2,-1, 0);
input_init(3,DEBOUNCE_DELAY,3,   PUSH_BUTTON,   3,-1, 0);
input_init(4,DEBOUNCE_DELAY,4,   PUSH_BUTTON,   4,-1, 0);
input_init(5,DEBOUNCE_DELAY,5,   PUSH_BUTTON,   5,-1, 0);
input_init(6,DEBOUNCE_DELAY,6,   PUSH_BUTTON,   6,-1, 0);
input_init(7,DEBOUNCE_DELAY,7,   PUSH_BUTTON,   7,-1, 0);
input_init(8,DEBOUNCE_DELAY,8,   PUSH_BUTTON,   8,-1, 0);
input_init(9,DEBOUNCE_DELAY,9,   PUSH_BUTTON,   9,-1, 0);
input_init(10,DEBOUNCE_DELAY,10, PUSH_BUTTON,   10,-1, 0);
input_init(11,DEBOUNCE_DELAY,14, PUSH_BUTTON,   11,-1, 0);

}

void discrete_task( void)
{
    int bit,flag;
    int bit_state;
    int i,itmp;

    for(i=0;i<BITS_MAXIMUM;i++)
    {
        if(bits[i].pulse_timer)
        {
            if(--bits[i].pulse_timer == 0)
            {
                switch(bits[i].mode)
                {
                    case PUSH_BUTTON:
                    
                    break;

                    case ENCODER_A:
                        if(bits[i+1].signal)
                        {
                            Joystick.releaseButton(bits[i].emitted_code_rising);
                        } else
                        {
                            Joystick.releaseButton(bits[i+1].emitted_code_rising);
                        }
                    break;

                    case ENCODER_B:
                    
                    break;

                    case TOGGLE_SWITCH:
                        Joystick.releaseButton(bits[i].emitted_code_rising);
                        Joystick.releaseButton(bits[i].emitted_code_falling);
                    break;

                }
            }
        }

        if(bits[i].flag & FLAG_ENABLED)
        {
            
            //if(bits[i].pin) // regular signal
                bit_state = digitalRead(bits[i].pin);
            //else // lever switch virtual signal
            //    bit_state = lever_state[i];

            flag = bits[i].flag;
            bit = (flag & FLAG_INVERTED) ^ bit_state;

            if(bit)
            {
                if(bits[i].counter < bits[i].max)
                {
                    if(++(bits[i].counter) >= bits[i].max)
                    {
                        if(bits[i].signal == 0)
                        {
                            //front __0__/--1-- 
                            switch(bits[i].mode)
                            {
                                case PUSH_BUTTON:
                                    if(i==4)
                                    {
                                       reverse_lever[0].state = REVERSE_LEVER_UP;
                                    }
                                    if(i==5)
                                    {
                                       reverse_lever[0].state = REVERSE_LEVER_DOWN;
                                    }
                                    if(i==8)
                                    {
                                       reverse_lever[1].state = REVERSE_LEVER_UP;
                                    }
                                    if(i==9)
                                    {
                                       reverse_lever[1].state = REVERSE_LEVER_DOWN;
                                    }
                                    Joystick.pressButton(bits[i].emitted_code_rising);
                                    Serial.write(STX);
                                    Serial.write('8');
                                    itmp = bits[i].emitted_code_rising;
                                    Serial.write(hex_table[itmp >> 4]);
                                    Serial.write(hex_table[itmp & 0x0F]);
                                    Serial.write(ETX);
                                break;

                                case ROTARY_SWITCH_POLE:
                                    Serial.write(STX);
                                    Serial.write('1');
                                    itmp = bits[i].emitted_code_rising;
                                    Serial.write(hex_table[itmp >> 4]);
                                    Serial.write(hex_table[itmp & 0x0F]);
                                    Serial.write(ETX);
                                break;

                                case ENCODER_A:
                                    if(bits[i+1].signal)
                                    {
                                        Joystick.pressButton(bits[i].emitted_code_rising);
                                    } else
                                    {
                                        Joystick.pressButton(bits[i+1].emitted_code_rising);
                                    }
                                break;

                                case ENCODER_B:
                                break;

                                case TOGGLE_SWITCH:
                                    bits[i].pulse_timer = bits[i].pulse_timer_value;
                                    Joystick.pressButton(bits[i].emitted_code_rising);
                                    //Joystick.releaseButton(bits[i].emitted_code_rising[0]);
                                    Joystick.pressButton(bits[i].emitted_code_rising);
                                    Serial.write(STX);
                                    Serial.write('8');
                                    itmp = bits[i].emitted_code_rising;
                                    Serial.write(hex_table[itmp >> 4]);
                                    Serial.write(hex_table[itmp & 0x0F]);
                                    Serial.write(ETX);
                                
                                break;

                                case SHIFT:
                                    //shift = bits[i].emitted_code_rising[0];
                                break;

                            }
                            bits[i].signal = 1;
                        }

                    }
                }
            }
            else
            {
                if(bits[i].counter > 0)
                {
                    if(--(bits[i].counter) <= 0)
                    {
                        if(bits[i].signal)
                        {
                            // edge  --1--\___0___ 
                            switch(bits[i].mode)
                            {
                                case PUSH_BUTTON:
                                    if(i==5 || i==4)
                                    {
                                       reverse_lever[0].state = REVERSE_LEVER_CENTER;
                                    }
                                    if(i==8 || i==9)
                                    {
                                       reverse_lever[1].state = REVERSE_LEVER_CENTER;
                                    }
                                    Joystick.releaseButton(bits[i].emitted_code_rising);
                                    Serial.write(STX);
                                    Serial.write('4');
                                    itmp = bits[i].emitted_code_rising;
                                    Serial.write(hex_table[itmp >> 4]);
                                    Serial.write(hex_table[itmp & 0x0F]);
                                    Serial.write(ETX);
                                break;

                                case ENCODER_A:
                                    Joystick.pressButton(bits[i].emitted_code_rising);
                                    Joystick.pressButton(bits[i+1].emitted_code_rising);
                                break;

                                case ENCODER_B:
                                
                                break;

                                case TOGGLE_SWITCH:
                                    bits[i].pulse_timer = bits[i].pulse_timer_value;
                                    //Joystick.releaseButton(bits[i].emitted_code_falling);
                                    Joystick.pressButton(bits[i].emitted_code_falling);
                                    Serial.write(STX);
                                    Serial.write('4');
                                    itmp = bits[i].emitted_code_rising;
                                    Serial.write(hex_table[itmp >> 4]);
                                    Serial.write(hex_table[itmp & 0x0F]);
                                    Serial.write(ETX);
                                break;

                                case SHIFT:
                                    //shift = bits[i].emitted_code_falling[0];
                                break;
                            }
                            bits[i].signal = 0;
                        }
                    }
                }
            }
        }
    }
}

/***************************************

   C O N S O L E   F U N C T I O N S

****************************************/


typedef int (*EXE_FPTR)(int argc, char *argv[]);

typedef struct _ROM_EXEC_TABLE 
{
  const char *fname;
  const EXE_FPTR func;
} ROM_EXEC_TABLE;

static int is_quotes(int ch)
{
   if(strchr("\'\"`",ch))
                return 1;
return 0;
}

static int is_delimiter(int ch)
{
   if(strchr(">=<|",ch))
                return 1;
return 0;
}

int is_whitespaces(int ch)
{
   if(strchr(" \t",ch))
                return 1;
return 0;
}




static int get_token_(char *wp, char *quote)
{
char *p = wp;
*quote = 0x00;

   if(is_quotes(*p))
   {
     *quote = *p++;
    while(*p != *quote  && *p != 0x00)
    p++;

    *p++ = 0x00;
      return p - wp;
   } else
   {
    do {
     p++;
    }
  while(!is_whitespaces(*p) && *p != 0x00);
   }
return p - wp;
}


static int get_argv(char *input_line, int argc_max, char *argv[])
{
char *w = input_line;
int cnt = 0;
char quote;
    while(1)
    {
     if(is_whitespaces(*w))
     {
       if(*w == 0)
       return cnt;

       *w++ = 0;
       continue;
     }
     if(*w != 0)
     {
       argv[cnt++] = w;
       w += get_token_(w,&quote);
       if(quote)
       {
         ++argv[cnt-1];
         quote = 0x00;
       }
        if(cnt >= argc_max)
                      return cnt;
     } else
       break;
    }
return cnt;
}

#if 1
int find_then_execute(void* exe_table, int argc, char *argv[]) {
  int i;
  EXE_FPTR func;
  ROM_EXEC_TABLE *exe_files = (ROM_EXEC_TABLE*) exe_table;
  for (i = 0; exe_files[i].fname[0]; i++) {
    if (0 == strcmp(exe_files[i].fname, argv[0])) {
      func = exe_files[i].func;
      if (func) {
        Serial.println("");
        return func(argc, argv);
      }

    }

  }
  if(argc > 0)
  {
    Serial.println("");
    Serial.print("Command ");
    Serial.print(argv[0]);
    Serial.println(" not found");
  }
  else
  {
    operation_flags = 0;  
  }
  return 1;
}
#endif

int split_all(char *input_line, int argc_max, char *argv[], const char *delimiters)
{
char *w = input_line;
int cnt = 0;
    argv[cnt++] = w;
    while(argc_max)
    {
     if(strchr(delimiters,*w))
     {
       if(*w == 0)
       return cnt;

       *w++ = 0;
         argv[cnt++] = w;
         if(cnt >= argc_max)
                      return cnt;

       continue;
     }
     w++;
    }
return cnt;
}


static char *trim(char *instr)
{
    int len = strlen(instr);
    int i;
    for(i=0; (i<len) && (instr[i]!=0x0D) && (instr[i]!=0x0A); i++);
    instr[i] = 0x00;
    for(i=0; (i<len) && (instr[i]==0x20); i++);
    return &instr[i];
}

static int show_version(int argc, char *argv[]) 
{
    Serial.println("Throttle quadrant for the twin engines jet rev.0.1");
  return 0;
}

static int show_raw(int argc, char *argv[]) 
{
  operation_flags |= OPERATION_FLAG_SHOW_RAW_VALUES;
  return 0;
}

static int show_scaled(int argc, char *argv[]) 
{
  operation_flags |= OPERATION_FLAG_SHOW_SCALED;
  return 0;
}

static int show_error(int argc, char *argv[]) 
{
  operation_flags |= OPERATION_FLAG_SHOW_ERROR;
  return 0;
}

static int show_clean(int argc, char *argv[]) 
{
  operation_flags |= OPERATION_FLAG_SHOW_CLEAN;
  return 0;
}

static int show_error_table(int argc, char *argv[]) 
{
  int i;
  for(i=0;i<128;i++)
  {
    Serial.print(i);
    Serial.print(" ");
    Serial.print(error_left[i]);
    Serial.print(" ");
    Serial.println(error_right[i]);
  }
  
  
  return 0;
}

static int clear_error_table(int argc, char *argv[]) 
{
  int i;
  for(i=0;i<128;i++)
  {
    error_left[i] = 0;
    error_right[i] = 0;
  }
  
  
  return 0;
}

static int set_initial_position(int argc, char *argv[]) 
{
  int i;
  for(i=0;i<4;i++)
    pos0[i] = pos_pull[i];
  factor_calculation();
  return 0;  
}

static int set_opposite_position(int argc, char *argv[]) 
{
  int i;
  for(i=0;i<4;i++)
    pos1[i] = pos_pull[i];
  factor_calculation();
  return 0;  
}

static int save_settings(int argc, char *argv[]) 
{
  save_config();
  return 0;  
}

void save_config(void)
{
  int i;
  int addr = 0x0010;
  unsigned char *p = (unsigned char*)pos0;
  for(i=0;i<sizeof(pos0);i++)
     EEPROM.write(addr++,*p++);
  
  p = (unsigned char*)pos1;
  for(i=0;i<sizeof(pos1);i++)
     EEPROM.write(addr++,*p++);

  p = (unsigned char*)error_left;
  for(i=0;i<sizeof(error_left);i++)
     EEPROM.write(addr++,*p++);

  p = (unsigned char*)error_right;
  for(i=0;i<sizeof(error_right);i++)
     EEPROM.write(addr++,*p++);  
}

void load_config(void)
{
  int i;
  int addr = 0x0010;
  unsigned char *p = (unsigned char*)pos0;
  for(i=0;i<sizeof(pos0);i++)
     *p++ = EEPROM.read(addr++);
  
  p = (unsigned char*)pos1;
  for(i=0;i<sizeof(pos1);i++)
     *p++ = EEPROM.read(addr++);

  p = (unsigned char*)error_left;
  for(i=0;i<sizeof(error_left);i++)
     *p++ = EEPROM.read(addr++);

  p = (unsigned char*)error_right;
  for(i=0;i<sizeof(error_right);i++)
     *p++ = EEPROM.read(addr++);
  
}


void command_com(char *buf)
{
const ROM_EXEC_TABLE exe_files[] = {
      {"version",show_version},
      {"show_raw",show_raw},
      {"sr",show_raw},
      {"show_scaled",show_scaled},
      {"ss",show_scaled},
      {"show_error",show_error},
      {"se",show_error},
      {"show_clean",show_clean},
      {"sc",show_clean},
      {"show_error_table",show_error_table},
      {"set",show_error_table},      
      {"clear_error_table",clear_error_table},
      {"cet",clear_error_table}, 
      {"set_initial_position",set_initial_position},
      {"sip",set_initial_position}, 
      {"set_opposite_position",set_opposite_position},
      {"sop",set_opposite_position}, 
      {"save",save_settings},
      { "", NULL } 
    };  
  char *myargv[10];
  memset(myargv,0,sizeof(myargv));
  int cnt = get_argv(trim(buf),10,myargv);
    if(0 == strcmp(myargv[0],"list"))
    {
      int i; 
      int len = 0;
      Serial.println("Avalilable commands:");
      Serial.print("list ");
      for (i = 0; exe_files[i].fname[0]; i++)
      {
          len += (strlen(exe_files[i].fname)+1);
          Serial.print(exe_files[i].fname);
          Serial.print(" ");
          if(len > 60)
          {
            len = 0;
            Serial.println("");
          }
      }    
      Serial.println("");
      Serial.print(">");
      return;
    }
  
    find_then_execute(exe_files,cnt,myargv);
    Serial.println("");
    Serial.print(">");
}



